/*
                                 Editor Window
 */

#ifndef EDITOR_H
#define EDITOR_H

#include <gtk/gtk.h>
#include "toolbar.h"
#include "statusbar.h"
#include "editclist.h"
#include "editortex.h"

/*
 *	Editor units types:
 */
#define EDITOR_UNITS_METRIC		0	/* Meters. */
#define EDITOR_UNITS_METRIC_ALT_FEET	1	/* Meters (altitude is in feet). */
#define EDITOR_UNITS_ENGLISH		2	/* Feet. */

/*
 *	Editor OPIDs:
 */
#define EDITOR_OPID_NEW			10
#define EDITOR_OPID_OPEN		11
#define EDITOR_OPID_SAVE		12
#define EDITOR_OPID_SAVE_AS		13
#define EDITOR_OPID_REVERT		14
#define EDITOR_OPID_CLOSE		30
#define EDITOR_OPID_EXIT		31

#define EDITOR_OPID_CUT			50
#define EDITOR_OPID_COPY		51
#define EDITOR_OPID_PASTE		52
#define EDITOR_OPID_NEW_OBJECT		60
#define EDITOR_OPID_GOTO_OBJECT		61
#define EDITOR_OPID_MOVE_TO_CURSOR	62
#define EDITOR_OPID_DELETE_OBJECT	63

#define EDITOR_OPID_REFRESH		100
#define EDITOR_OPID_REFRESH_ALL		101
#define EDITOR_OPID_STOP		102

#define EDITOR_OPID_CHECK_SCENERY	110
#define EDITOR_OPID_RUN_SAR		120


/*
 *	Editor view display flags:
 */
#define EDITOR_VIEW_DISPLAY_MAP			(1 << 0)
#define EDITOR_VIEW_DISPLAY_OBJECT_LABELS	(1 << 1)
#define EDITOR_VIEW_DISPLAY_POSITION_LABEL	(1 << 2)
#define EDITOR_VIEW_DISPLAY_REGISTERED_LOCATIONS	(1 << 3)


/*
 *	Editor OPID structure:
 */
typedef struct {

	gpointer	editor;
	gint		id;		/* OPID. */
	gchar		*tooltip;
	void		(*func_cb)(GtkWidget *, gpointer);

} editor_opid_struct;

#define EDITOR_OPID(p)	((editor_opid_struct *)(p))


/*
 *	Editor window structure:
 */
typedef struct {

        gboolean	map_state;
        gint		busy_count;
        GtkAccelGroup	*accelgrp;
        gpointer	core_ptr;
	gboolean	processing;
	gint		page_num;
	gint		stop_count;
	gint		drag_button;
	guint		drag_key_state;
	gint		drag_last_x,
			drag_last_y;

        GtkWidget	*toplevel,
			*main_vbox,
			*menu_bar,
			*notebook;
	toolbar_struct	*tool_bar;

	editor_opid_struct **opid;
	gint total_opids;

	GdkPixmap	*list_object_pixmap;
	GdkBitmap	*list_object_mask;
        GdkPixmap       *list_ground_pixmap;
        GdkBitmap       *list_ground_mask;
        GdkPixmap       *list_helipad_pixmap;
        GdkBitmap       *list_helipad_mask;
        GdkPixmap       *list_runway_pixmap;
        GdkBitmap       *list_runway_mask;
        GdkPixmap       *list_human_pixmap;
        GdkBitmap       *list_human_mask;
        GdkPixmap       *list_smoke_pixmap;
        GdkBitmap       *list_smoke_mask;
        GdkPixmap       *list_fire_pixmap;
        GdkBitmap       *list_fire_mask;

	GdkBitmap	*view_object_mask,
			*view_ground_mask,
			*view_helipad_mask,
			*view_human_mask,
			*view_smoke_mask,
			*view_fire_mask;

	GtkWidget	*new_mi,		/* File. */
			*open_mi,
			*save_mi,
			*save_as_mi,
			*exit_mi,
			*cut_mi,		/* Edit. */
			*copy_mi,
			*paste_mi,
			*new_object_mi,
			*goto_selected_object_mi,
			*move_object_to_cursor_mi,
			*delete_object_mi,
			*view_header_micheck,	/* View. */
			*view_scenery_micheck,
			*view_map_micheck,
			*view_object_labels_micheck,
			*view_position_label_micheck,
			*view_registered_locations_micheck,
			*refresh_mi,
			*check_scenery_mi,
			*run_sar_mi;

	GtkWidget	*hdr_sw, *hdr_sw_eb,
			*hdr_name_entry,
			*hdr_gps_longitude_spin,
			*hdr_gps_latitude_spin,
			*hdr_planetary_radius_spin,
			*hdr_cant_spin,
			*hdr_elevation_spin,
			*hdr_ground_is_water_tb,
			*hdr_ground_tile_width_spin,
                        *hdr_ground_tile_height_spin,
                        *hdr_ground_tile_close_range_spin,
			*hdr_ground_color_btn,
			*hdr_ground_texture_entry,
			*hdr_map_da,
			*hdr_map_width_spin,
			*hdr_map_height_spin;
	editclist_struct	*hdr_textures_eclist,
				*hdr_registered_locations_eclist;
	editor_tex_struct	*hdr_map_tex;
	guint8		*hdr_map_buf;		/* 32 bits. */
	gint		hdr_freeze_count;

	GtkWidget	*scenery_vpaned, *view_hpaned,
			*view_da,
			*view_units_combo,
			*view_x_spin,
			*view_y_spin,
			*view_zoom_spin,
			*view_grid_spacing_spin,
			*view_cursor_x_spin,
                        *view_cursor_y_spin,
                        *view_cursor_z_spin,
			*obj_clist;
	GdkPixmap	*view_pm;
	guint8		*view_buf;		/* 32 bits. */
	gint		view_freeze_count;
	gulong		view_display_flags;
	gint		view_cursor_xw, view_cursor_yw;
	guint		view_cursor_draw_toid;
	guint		view_scroll_toid;
	gboolean	view_rect_select;
	gint		view_rect_select_x0, view_rect_select_y0,
			view_rect_select_x1, view_rect_select_y1;

	GtkWidget	*obj_sw, *obj_sw_eb, *obj_main_box,
			*obj_icon, *obj_name_entry, *obj_type_combo,
			*obj_model_entry,
			*obj_x_spin, *obj_y_spin, *obj_z_spin,
			*obj_heading_spin, *obj_pitch_spin, *obj_bank_spin,
			*obj_range_spin, *obj_range_far_spin,
			*obj_no_depth_test_tb,
			*obj_polygon_offset_tb,
			*obj_polygon_offset_reverse_tb,
			*obj_polygon_offset_write_depth_tb,
			*obj_size_shape_combo,
			*obj_size_radius_spin,
			*obj_size_x_min_spin, *obj_size_x_max_spin,
                        *obj_size_y_min_spin, *obj_size_y_max_spin,
                        *obj_size_z_min_spin, *obj_size_z_max_spin,
			*obj_crash_type_combo,
			*obj_support_surface_tb,
			*obj_ref_name_entry, *obj_ref_change_btn,
			*obj_ref_x_spin, *obj_ref_y_spin, *obj_ref_z_spin,
			*obj_ref_heading_spin, *obj_ref_pitch_spin, *obj_ref_bank_spin;

	GtkWidget	*obj_menu,
			*obj_cut_mi,
                        *obj_copy_mi,
                        *obj_paste_mi,
                        *obj_new_object_mi,
                        *obj_goto_selected_object_mi,
                        *obj_move_object_to_cursor_mi,
                        *obj_delete_object_mi,
			*obj_view_map_micheck,
			*obj_view_object_labels_micheck,
			*obj_view_position_label_micheck,
			*obj_view_registered_locations_micheck;
	gint		obj_freeze_count;
	gint		obj_selected;	/* Current selected object. */
	gint		obj_sw_button;
	gint		obj_sw_button_last_x, obj_sw_button_last_y;

	GtkWidget	*obj_helipad_frame,
			*obj_helipad_type_combo,
			*obj_helipad_edge_lighting_tb,
			*obj_helipad_fuel_tb,
			*obj_helipad_repair_tb,
			*obj_helipad_dropoff_tb,
			*obj_helipad_restart_point_tb,
			*obj_helipad_length_spin,
			*obj_helipad_width_spin,
			*obj_helipad_recession_spin,
			*obj_helipad_label_entry;

	GtkWidget	*obj_runway_frame,
			*obj_runway_surface_combo,
			*obj_runway_length_spin,
			*obj_runway_width_spin,
			*obj_runway_dashes_spin,
			*obj_runway_edge_light_spacing_spin,
			*obj_runway_thresholds_tb,
			*obj_runway_borders_tb,
			*obj_runway_td_markers_tb,
			*obj_runway_midway_markers_tb,
			*obj_runway_north_gs_tb,
			*obj_runway_south_gs_tb,
			*obj_runway_north_label_entry,
			*obj_runway_south_label_entry,
			*obj_runway_north_displaced_threshold_spin,
			*obj_runway_south_displaced_threshold_spin,
			*obj_runway_north_lighting_end_tb,
                        *obj_runway_north_lighting_tracer_tb,
                        *obj_runway_north_lighting_align_tb,
                        *obj_runway_north_lighting_ils_glide_tb,
                        *obj_runway_south_lighting_end_tb,
                        *obj_runway_south_lighting_tracer_tb,
                        *obj_runway_south_lighting_align_tb,
                        *obj_runway_south_lighting_ils_glide_tb;

        GtkWidget       *obj_human_frame,
                        *obj_human_type_combo,
                        *obj_human_need_rescue_tb,
                        *obj_human_sit_tb,
                        *obj_human_sit_down_tb,
                        *obj_human_sit_up_tb,
                        *obj_human_lying_tb,
                        *obj_human_alert_tb,
                        *obj_human_aware_tb,
                        *obj_human_in_water_tb,
                        *obj_human_on_streatcher_tb,
                        *obj_human_run_tb,
                        *obj_human_run_towards_tb,
                        *obj_human_run_away_tb,
			*obj_human_message_enter_entry;

	GtkWidget	*obj_preset_model_frame,
			*obj_preset_model_type_combo,
			*obj_preset_model_values_entry;

	status_bar_struct	*status_bar;

	gint		units;		/* One of EDITOR_UNITS_*. */
	gboolean	has_changes;
	gchar		*scn_filename;
	gchar		*scn_title;
	gchar		*scn_description;



} editor_struct;

#define EDITOR(p)	((editor_struct *)(p))
#define EDITOR_IS_MAPPED(p)	(((p) != NULL) ? \
 (EDITOR(p)->map_state) : FALSE)
#define EDITOR_TOPLEVEL(p)	(((p) != NULL) ? \
 (EDITOR(p)->toplevel) : NULL)
#define EDITOR_TOOL_BAR(p)	(((p) != NULL) ? \
 (EDITOR(p)->tool_bar) : NULL)
#define EDITOR_NOTEBOOK(p)	(((p) != NULL) ? \
 (EDITOR(p)->notebook) : NULL)
#define EDITOR_STATUS_BAR(p)	(((p) != NULL) ? \
 (EDITOR(p)->status_bar) : NULL)

#define EDITOR_PROCESSING(p)	(((p) != NULL) ? \
 (EDITOR(p)->processing) : FALSE)
#define EDITOR_PAGE_NUMBER(p)	(((p) != NULL) ? \
 (EDITOR(p)->page_num) : 0)
#define EDITOR_STOP_COUNT(p)	(((p) != NULL) ? \
 (EDITOR(p)->stop_count) : 0)
#define EDITOR_UNITS(p)		(((p) != NULL) ? \
 (EDITOR(p)->units) : EDITOR_UNITS_METRIC)
#define EDITOR_HAS_CHANGES(p)	(((p) != NULL) ? \
 (EDITOR(p)->has_changes) : FALSE)


extern gint EditorGetObjType(GtkWidget *w);
extern void EditorSetObjType(GtkWidget *w, gint type);
extern gint EditorGetSizeShape(GtkWidget *w);
extern void EditorSetSizeShape(GtkWidget *w, gint shape);
extern gint EditorGetRunwaySurface(GtkWidget *w);
extern void EditorSetRunwaySurface(GtkWidget *w, gint surface);
extern gboolean EditorIsObjectNameSpecial(const gchar *name);
extern gboolean EditorGetRegisteredLocation(
        editor_struct *e, gint i,
        gfloat *xm, gfloat *ym, gfloat *zm,
        gfloat *heading, gfloat *pitch, gfloat *bank,
        gchar **name
);

extern void EditorHdrMapDraw(editor_struct *e);

extern editor_struct *EditorNew(gpointer core_ptr);
extern void EditorUpdateTitle(editor_struct *e);
extern void EditorUpdateMenus(editor_struct *e);
extern void EditorSavePosition(editor_struct *e);
extern void EditorSetBusy(editor_struct *e, gboolean is_busy);
extern void EditorStopReset(editor_struct *e, gboolean sensitive);
extern void EditorMap(editor_struct *e);
extern void EditorUnmap(editor_struct *e);
extern void EditorDelete(editor_struct *e);

#endif	/* EDITOR_H */
