#include "obj.h"
#include <gtk/gtk.h>
#include "editorobj.h"


editor_obj_struct *EditorObjNew(void);
editor_obj_struct *EditorObjNewWithValues(
        gint type,
        const sar_position_struct *pos,
        const sar_direction_struct *dir,
        const sar_contact_bounds_struct *cb,
        const gchar *name,
        const gchar *model_file
);
editor_obj_struct *EditorObjDup(const editor_obj_struct *o);
void EditorObjDelete(editor_obj_struct *o);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))
#define IS_STRING_EMPTY(s)      (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define RADTODEG(r)     ((r) * 180 / PI)
#define DEGTORAD(d)     ((d) * PI / 180)


/*
 *	Creates a new object with values reset.
 */
editor_obj_struct *EditorObjNew(void)
{
	editor_obj_struct *o = EDITOR_OBJ(
	    g_malloc0(sizeof(editor_obj_struct))
	);
	if(o == NULL)
	    return(NULL);

	/* Reset values that are not to be initialized as 0's. */
	o->type = SAR_OBJ_TYPE_STATIC;
	o->ref_obj_num = -1;
	o->smoke_radius_rate = -1.0f;

	return(o);
}

/*
 *	Creates a new object with values.
 */
editor_obj_struct *EditorObjNewWithValues(
        gint type,
        const sar_position_struct *pos,
        const sar_direction_struct *dir,
        const sar_contact_bounds_struct *cb,
        const gchar *name,
        const gchar *model_file
)
{
	editor_obj_struct *o = EditorObjNew();

	o->type = type;
	if(pos != NULL)
	    memcpy(&o->pos, pos, sizeof(sar_position_struct));
	if(dir != NULL)
	    memcpy(&o->dir, dir, sizeof(sar_direction_struct));
	if(cb != NULL)
	    memcpy(&o->cb, cb, sizeof(sar_contact_bounds_struct));
	o->name = STRDUP(name);
	o->model_file = STRDUP(model_file);

	return(o);
}

/*
 *	Coppies the object.
 */
editor_obj_struct *EditorObjDup(const editor_obj_struct *o)
{
	editor_obj_struct *o2;

	if(o == NULL)
	    return(NULL);

	o2 = EditorObjNew();
        if(o2 == NULL)
            return(NULL);

	/* Begin copying values from source object o to the new
	 * target object o2.
	 */

	/* Basics. */
	o2->type = o->type;
	o2->flags = o->flags;
	memcpy(&o2->pos, &o->pos, sizeof(sar_position_struct));
	memcpy(&o2->dir, &o->dir, sizeof(sar_direction_struct));
	memcpy(&o2->cb, &o->cb, sizeof(sar_contact_bounds_struct));
	o2->range = o->range;
	o2->range_far = o->range_far;
	o2->name = STRDUP(o->name);
        o2->description = STRDUP(o->description);
        o2->model_file = STRDUP(o->model_file);

	/* Reference object. */
        o2->ref_obj_name = STRDUP(o->ref_obj_name);
        o2->ref_obj_num = o->ref_obj_num;
        memcpy(&o2->ref_offset, &o->ref_offset, sizeof(sar_position_struct));
        memcpy(&o2->ref_dir, &o->ref_dir, sizeof(sar_direction_struct));

        /* Helipad. */
	o2->helipad_flags = o->helipad_flags;
	o2->helipad_type_name = STRDUP(o->helipad_type_name);
	o2->helipad_length = o->helipad_length;
        o2->helipad_width = o->helipad_width;
        o2->helipad_recession = o->helipad_recession;
	o2->helipad_label = STRDUP(o->helipad_label);

	/* Runway. */
	o2->runway_flags = o->runway_flags;
        o2->runway_length = o->runway_length;
        o2->runway_width = o->runway_width;
	o2->runway_surface = o->runway_surface;
	o2->runway_dashes = o->runway_dashes;
	o2->runway_edge_light_spacing = o->runway_edge_light_spacing;
        o2->runway_north_label = STRDUP(o->runway_north_label);
        o2->runway_south_label = STRDUP(o->runway_south_label);
	o2->runway_north_displaced_threshold = o->runway_north_displaced_threshold;
	o2->runway_south_displaced_threshold = o->runway_south_displaced_threshold;
	o2->runway_north_lighting_flags = o->runway_north_lighting_flags;
	o2->runway_south_lighting_flags = o->runway_south_lighting_flags;

	/* Human. */
        o2->human_flags = o->human_flags;
        o2->human_type_name = STRDUP(o->human_type_name);
	o2->human_message_enter = STRDUP(o->human_message_enter);

	/* Fire. */
        o2->fire_radius = o->fire_radius;
        o2->fire_height = o->fire_height;

	/* Smoke. */
	memcpy(&o2->smoke_offset, &o->smoke_offset, sizeof(sar_position_struct));
        o2->smoke_radius_start = o->smoke_radius_start;
        o2->smoke_radius_max = o->smoke_radius_max;
        o2->smoke_radius_rate = o->smoke_radius_rate;
        o2->smoke_hide_at_max = o->smoke_hide_at_max;
        o2->smoke_respawn_int = o->smoke_respawn_int;
        o2->smoke_total_units = o->smoke_total_units;
        o2->smoke_color_code = o->smoke_color_code;

	/* Premodeled. */
        o2->premodeled_type_name = STRDUP(o->premodeled_type_name);
        o2->premodeled_value = STRDUP(o->premodeled_value);

	return(o2);
}

/*
 *	Deletes the object.
 */
void EditorObjDelete(editor_obj_struct *o)
{
	if(o == NULL)
	    return;

	g_free(o->name);
	g_free(o->description);
	g_free(o->model_file);

	g_free(o->ref_obj_name);

        g_free(o->helipad_type_name);
        g_free(o->helipad_label);

        g_free(o->runway_north_label);
        g_free(o->runway_south_label);

        g_free(o->human_type_name);
	g_free(o->human_message_enter);

        g_free(o->premodeled_type_name);
        g_free(o->premodeled_value);

	g_free(o);
}

