#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gtk/gtk.h>

#include "matrixmath.h"
#include "blitting.h"
#include "sfm.h"	/* For SFM math functions. */
#include "editor.h"
#include "editortex.h"
#include "editorobj.h"
#include "editorlist.h"
#include "editorview.h"
#include "scnedit.h"
#include "config.h"


void EditorViewMToW(
        gfloat xm, gfloat ym,
        gfloat offset_xm, gfloat offset_ym, gfloat zoom,
        gint *xw, gint *yw,
        gint widthw, gint heightw
);
void EditorViewGetViewPositionM(
        editor_struct *e, gfloat *x, gfloat *y
);
void EditorViewSetViewPositionM(
        editor_struct *e, gfloat x, gfloat y
);
void EditorViewGetViewZoomM(
        editor_struct *e, gfloat *d
);
void EditorViewSetViewZoomM(
        editor_struct *e, gfloat d
);
void EditorViewGetViewGridSpacingM(
        editor_struct *e, gfloat *d
);
void EditorViewSetViewGridSpacingM(
        editor_struct *e, gfloat d
);
void EditorViewGetPositionWToM(
        editor_struct *e, gint wx, gint wy,
        gfloat *x, gfloat *y
);
void EditorViewGetCursorM(
        editor_struct *e, gfloat *x, gfloat *y, gfloat *z
);
void EditorViewSetCursorM(
        editor_struct *e, gfloat x, gfloat y, gfloat z
);

static void EditorViewDrawBackground(
	editor_struct *e, GdkPixmap *pixmap, gulong display_flags,
	gint width, gint height, gint half_width, gint half_height,
	gint state, GtkStyle *style,
	gfloat offset_xm, gfloat offset_ym, gfloat zoom
);
static void EditorViewDrawObject(
	editor_struct *e, GdkPixmap *pixmap, gulong display_flags,
	gint width, gint height, gint half_width, gint half_height,
	gint state, GtkStyle *style, GdkGC *gc,
	gfloat offset_xm, gfloat offset_ym, gfloat zoom,
	gint on, editor_obj_struct *o
);
void EditorViewDraw(editor_struct *e);
gint EditorViewCursorDrawTOCB(gpointer data);

gint EditorViewProcessRectangularSelect(editor_struct *e);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))
#define IS_STRING_EMPTY(s)      (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define RADTODEG(r)     ((r) * 180 / PI)
#define DEGTORAD(d)     ((d) * PI / 180)


/*
 *	Returns the given position in meters as window coordinates.
 */
void EditorViewMToW(
	gfloat xm, gfloat ym,
	gfloat offset_xm, gfloat offset_ym, gfloat zoom,
        gint *xw, gint *yw,
	gint widthw, gint heightw
)
{
	if(zoom <= 0.0f)
	{
	    if(xw != NULL)
		*xw = 0;
            if(yw != NULL)
                *yw = 0;
	    return;
	}

	if(xw != NULL)
	    *xw = (widthw / 2) +
		(gint)((xm - offset_xm) * widthw / zoom);
	if(yw != NULL)
	    *yw = (heightw / 2) -
		(gint)((ym - offset_ym) * heightw / zoom);
}

/*
 *	Returns the current view position in meters.
 */
void EditorViewGetViewPositionM(
        editor_struct *e, gfloat *x, gfloat *y
)
{
	GtkAdjustment *adj;
	GtkWidget *w;


	if(e == NULL)
	    return;

	/* Get values in displayed units. */
	w = e->view_x_spin;
	adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
	if((adj != NULL) && (x != NULL))
	    *x = adj->value;
        w = e->view_y_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if((adj != NULL) && (y != NULL))
            *y = adj->value;

        /* Convert value as needed. */
	switch(EDITOR_UNITS(e))
	{
	  case EDITOR_UNITS_METRIC:
          case EDITOR_UNITS_METRIC_ALT_FEET:
	    break;
	  case EDITOR_UNITS_ENGLISH:
	    if(x != NULL)
		*x = (gfloat)SFMFeetToMeters(*x);
            if(y != NULL)
                *y = (gfloat)SFMFeetToMeters(*y);
	    break;
	}
}

/*
 *	Sets the view position in meters and redraws the view.
 */
void EditorViewSetViewPositionM(
        editor_struct *e, gfloat x, gfloat y
)
{
	GtkAdjustment *adj;
	GtkWidget *w;


	if(e == NULL)
	    return;

        /* Convert value as needed. */
        switch(EDITOR_UNITS(e))
        {
          case EDITOR_UNITS_METRIC:
            break;
          case EDITOR_UNITS_METRIC_ALT_FEET:
            break;
          case EDITOR_UNITS_ENGLISH:
            x = (gfloat)SFMMetersToFeet(x);
            y = (gfloat)SFMMetersToFeet(y);
            break;
        }

        e->view_freeze_count++;

        /* Update displayed widget values. */
        w = e->view_x_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if(adj != NULL)
        {
            adj->value = x;
            gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
        }

        w = e->view_y_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if(adj != NULL)
        {
            adj->value = y;
            gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
        }

	e->view_freeze_count--;
	EditorViewDraw(e);
}

/*
 *	Returns the zoom distance in meters.
 */
void EditorViewGetViewZoomM(
        editor_struct *e, gfloat *d
)
{
        GtkAdjustment *adj;
        GtkWidget *w;


        if(e == NULL)
            return;

        /* Get value in displayed units. */
        w = e->view_zoom_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if((adj != NULL) && (d != NULL))
            *d = adj->value;

	/* Convert value as needed. */
        switch(EDITOR_UNITS(e))
        {
          case EDITOR_UNITS_METRIC:
          case EDITOR_UNITS_METRIC_ALT_FEET:
            break;
          case EDITOR_UNITS_ENGLISH:
            if(d != NULL)
                *d = (gfloat)SFMFeetToMeters(*d);
            break;
        }
}

/*
 *	Sets the view zoom in meters and redraws the view.
 */
void EditorViewSetViewZoomM(
        editor_struct *e, gfloat d
)
{
        GtkAdjustment *adj;
        GtkWidget *w;


        if((e == NULL) || (d <= 0.0f))
            return;

        /* Convert value as needed. */
        switch(EDITOR_UNITS(e))
        {
          case EDITOR_UNITS_METRIC:
            break;
          case EDITOR_UNITS_METRIC_ALT_FEET:
            break;
          case EDITOR_UNITS_ENGLISH:
            d = (gfloat)SFMMetersToFeet(d);
            break;
        }

        e->view_freeze_count++;

        /* Update displayed widget values. */
        w = e->view_zoom_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if(adj != NULL)
        {
            adj->value = d;
            gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
        }

        e->view_freeze_count--;
        EditorViewDraw(e);
}

/*
 *      Returns the grid spacing in meters.
 */
void EditorViewGetViewGridSpacingM(
        editor_struct *e, gfloat *d
)
{
        GtkAdjustment *adj;
        GtkWidget *w;


        if(e == NULL)
            return;

        /* Get value in displayed units. */
        w = e->view_grid_spacing_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if((adj != NULL) && (d != NULL))
            *d = adj->value;

        /* Convert value as needed. */
        switch(EDITOR_UNITS(e))
        {
          case EDITOR_UNITS_METRIC:
          case EDITOR_UNITS_METRIC_ALT_FEET:
            break;
          case EDITOR_UNITS_ENGLISH:
            if(d != NULL)
                *d = (gfloat)SFMFeetToMeters(*d);
            break;
        }
}

/*
 *	Sets the view's grid spacing and redraws the view.
 */
void EditorViewSetViewGridSpacingM(
        editor_struct *e, gfloat d
)
{
        GtkAdjustment *adj;
        GtkWidget *w;


        if((e == NULL) || (d <= 0.0f))
            return;

        /* Convert value as needed. */
        switch(EDITOR_UNITS(e))
        {
          case EDITOR_UNITS_METRIC:
            break;
          case EDITOR_UNITS_METRIC_ALT_FEET:
            break;
          case EDITOR_UNITS_ENGLISH:
            d = (gfloat)SFMMetersToFeet(d);
            break;
        }

        e->view_freeze_count++;

        /* Update displayed widget values. */
        w = e->view_grid_spacing_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if(adj != NULL)
        {
            adj->value = d;
            gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
        }

        e->view_freeze_count--;
        EditorViewDraw(e);
}

/*
 *	Get position in meters from window coordinates.
 */
void EditorViewGetPositionWToM(
        editor_struct *e, gint wx, gint wy,
        gfloat *x, gfloat *y
)
{
	GtkWidget *w;
	gint wwidth, wheight;
	gfloat zoom, offset_xm, offset_ym;

	if(x != NULL)
	    *x = 0.0f;
        if(y != NULL)
            *y = 0.0f;

        if(e == NULL)
            return;

	/* Get zoom in meters. */
	EditorViewGetViewZoomM(e, &zoom);

	/* Get view window and its size. */
	w = e->view_da;
	if(w == NULL)
	    return;

	wwidth = w->allocation.width;
        wheight = w->allocation.height;
	if((wwidth <= 0) || (wheight <= 0))
	    return;

        /* Get offset in meters. */
        EditorViewGetViewPositionM(e, &offset_xm, &offset_ym);

	if(x != NULL)
	    *x = ((gfloat)(wx - (wwidth / 2)) * zoom / (gfloat)wwidth) +
		offset_xm;
	if(y != NULL)
	    *y = ((gfloat)((wheight / 2) - wy) * zoom / (gfloat)wheight) +
		offset_ym;
}


/*
 *	Returns the cursor position in meters.
 */
void EditorViewGetCursorM(
        editor_struct *e, gfloat *x, gfloat *y, gfloat *z
)
{
        GtkAdjustment *adj;
        GtkWidget *w;


        if(x != NULL)
            *x = 0.0f;
        if(y != NULL)
            *y = 0.0f;
        if(z != NULL)
            *z = 0.0f;

        if(e == NULL)
            return;

        /* Get value in displayed units. */
        w = e->view_cursor_x_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if((adj != NULL) && (x != NULL))
            *x = adj->value;
        w = e->view_cursor_y_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if((adj != NULL) && (y != NULL))
            *y = adj->value;
        w = e->view_cursor_z_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if((adj != NULL) && (z != NULL))
            *z = adj->value;

        /* Convert value as needed. */
        switch(EDITOR_UNITS(e))
        {
          case EDITOR_UNITS_METRIC:
            break;
	  case EDITOR_UNITS_METRIC_ALT_FEET:
            if(z != NULL)
                *z = (gfloat)SFMFeetToMeters(*z);
	    break;
          case EDITOR_UNITS_ENGLISH:
            if(x != NULL)
                *x = (gfloat)SFMFeetToMeters(*x);
            if(y != NULL)
                *y = (gfloat)SFMFeetToMeters(*y);
            if(z != NULL)
                *z = (gfloat)SFMFeetToMeters(*z);
            break;
        }
}

/*
 *      Sets the cursor position in meters and redraws the view.
 */
void EditorViewSetCursorM(
        editor_struct *e, gfloat x, gfloat y, gfloat z
)
{
        GtkAdjustment *adj;
        GtkWidget *w;


        if(e == NULL)
            return;

        /* Convert value as needed. */
        switch(EDITOR_UNITS(e))
        {
          case EDITOR_UNITS_METRIC:
            break;
          case EDITOR_UNITS_METRIC_ALT_FEET:
            z = (gfloat)SFMMetersToFeet(z);
            break;
          case EDITOR_UNITS_ENGLISH:
            x = (gfloat)SFMMetersToFeet(x);
            y = (gfloat)SFMMetersToFeet(y);
            z = (gfloat)SFMMetersToFeet(z);
            break;
        }

        e->view_freeze_count++;

	/* Update displayed widget values. */
        w = e->view_cursor_x_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if(adj != NULL)
	{
            adj->value = x;
	    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
	}

        w = e->view_cursor_y_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if(adj != NULL)
        {
            adj->value = y;
            gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
        }

        w = e->view_cursor_z_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if(adj != NULL)
        {
            adj->value = z;
            gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
        }

        e->view_freeze_count--;
	EditorViewDraw(e);
}


/*
 *      Called by EditorViewDraw() to draw the map.
 */
static void EditorViewDrawBackground(
        editor_struct *e, GdkPixmap *pixmap, gulong display_flags,
        gint width, gint height, gint half_width, gint half_height,
        gint state, GtkStyle *style,
        gfloat offset_xm, gfloat offset_ym, gfloat zoom
)
{
	gint src_xw, src_yw, tar_xw, tar_yw;
	gfloat src_zoom_x, src_zoom_y;
	GtkAdjustment *adj;
	GtkSpinButton *spin;
	gfloat map_widthm, map_heightm;
	const editor_tex_struct *t = e->hdr_map_tex;
	guint8 *tar_buf = e->view_buf;


#define DO_DRAW_SOLID_BG			\
{						\
 gdk_draw_rectangle(				\
  pixmap, style->base_gc[state], TRUE,		\
  0, 0, width, height				\
 );						\
}
	/* Just draw solid background? */
	if(!(display_flags & EDITOR_VIEW_DISPLAY_MAP))
	{
            DO_DRAW_SOLID_BG
            return;
        }

	/* No map loaded or view image buffer not allocated? */
	if((t == NULL) || (tar_buf == NULL))
	{
	    DO_DRAW_SOLID_BG
	    return;
	}

	/* Get map size in meters. */
	spin = (GtkSpinButton *)e->hdr_map_width_spin;
	adj = (spin != NULL) ? spin->adjustment : NULL;
	map_widthm = (adj != NULL) ? adj->value : 1000.0f;
        spin = (GtkSpinButton *)e->hdr_map_height_spin;
        adj = (spin != NULL) ? spin->adjustment : NULL;
        map_heightm = (adj != NULL) ? adj->value : 1000.0f;
	if((map_widthm <= 0.0f) || (map_heightm <= 0.0f))
	{
	    DO_DRAW_SOLID_BG
	    return;
	}

	/* From here on, the map texture data is considered the source
	 * and the pixmap is considered the target.
	 */

        /* Calculate the zoom coefficient needed to draw the source to
	 * its proper size on the target.
	 */
        src_zoom_x = map_widthm / zoom * width / t->width;
        src_zoom_y = map_heightm / zoom * height / t->height;
	if((src_zoom_x <= 0.0f) || (src_zoom_y <= 0.0f))
	{
            DO_DRAW_SOLID_BG
            return;
        }

	/* Calculate the position of the source to be drawn on the
	 * target in window coordinates.
	 */
	tar_xw = (gint)(
	    half_width - (map_widthm / zoom * half_width) -
	    (offset_xm / zoom * width)
	);
	tar_yw = (gint)(
	    half_height - (map_heightm / zoom * half_height) +
	    (offset_ym / zoom * height)
	);

	/* Calculate the starting position within the source. */
        src_xw = (gint)(-tar_xw / src_zoom_x);
	src_yw = (gint)(-tar_yw / src_zoom_y);


	/* Source small enough to draw? */
	if((src_zoom_x <= 20.0f) && (src_zoom_y <= 20.0f))
	{
	    BlitClearRGBA(
		BLIT_PIXEL_OP_SET,
		(guint32 *)tar_buf, width, height, -1,
		0xffffffff
	    );
	    BlitImageScaleRGBA(
		BLIT_PIXEL_OP_SET,
		(guint32 *)tar_buf,
		tar_xw, tar_yw, width, height, -1,
		(const guint32 *)t->data,
		src_xw, src_yw, t->width, t->height, t->bpl,
		src_zoom_x, src_zoom_y
	    );
	    gdk_draw_rgb_32_image(
		pixmap, style->white_gc,
		0, 0, width, height,
		GDK_RGB_DITHER_NORMAL,
		(guchar *)tar_buf,
		width * 4
	    );
	}
	else
	{
	    DO_DRAW_SOLID_BG
	}
#undef DO_DRAW_SOLID_BG
}

/* 
 *	Called by EditorViewDraw() to draw an object.
 */
static void EditorViewDrawObject(
        editor_struct *e, GdkPixmap *pixmap, gulong display_flags,
        gint width, gint height, gint half_width, gint half_height,
	gint state, GtkStyle *style, GdkGC *gc,
        gfloat offset_xm, gfloat offset_ym, gfloat zoom,
	gint on, editor_obj_struct *o
)
{
	gint x, y;
	gfloat xm, ym;
	const sar_position_struct *pos;
	const sar_direction_struct *dir, *dir2 = NULL;
	const sar_contact_bounds_struct *cb;
	GdkBitmap *mask;


	if(o == NULL)
	    return;

        /* Get gc for selected state as needed. */
        if(state == GTK_STATE_SELECTED)
            gc = style->bg_gc[state];

	pos = &o->pos;
	dir = &o->dir;
	if(o->ref_obj_num > -1)
	{
	    const editor_obj_struct *o2 = EDITOR_OBJ(
		EditorObjListGetObj(e, o->ref_obj_num)
	    );
	    if((o2 != NULL) && (o2 != o))
		dir2 = &o2->dir;
	}
	xm = (gfloat)pos->x;
	ym = (gfloat)pos->y;

	/* Calculate center of object in window coordinates. */
	EditorViewMToW(
	    xm, ym,
	    offset_xm, offset_ym, zoom,
	    &x, &y,
	    width, height
	);

	/* Draw contact bounds. */
	cb = &o->cb;
	if(TRUE)
	{
	    gint rxw, ryw;
	    gdouble a[4], r[4], theta;
	    GdkPoint pt[5];

	    switch(cb->contact_shape)
	    {
	      case SAR_CONTACT_SHAPE_SPHERICAL:
	      case SAR_CONTACT_SHAPE_CYLENDRICAL:
		if((gfloat)cb->contact_radius <= 0.0f)
		    break;
		rxw = (gint)(cb->contact_radius * width / zoom);
		ryw = (gint)(cb->contact_radius * height / zoom);
                gdk_draw_arc(
                    pixmap, gc, FALSE,
		    x - rxw, y - ryw,
		    2 * rxw, 2 * ryw,
		    0 * 64, 360 * 64
		);
		break;

	      case SAR_CONTACT_SHAPE_RECTANGULAR:
		theta = ((dir2 != NULL) ? dir2->heading : 0.0 * PI) +
		    dir->heading;

		a[0] = cb->contact_x_min;
                a[1] = cb->contact_y_max;
		a[2] = 0.0;
		MatrixRotateHeading3(a, theta, r);
		pt[0].x = (gint16)(x + (gint)(r[0] * width / zoom));
		pt[0].y = (gint16)(y - (gint)(r[1] * height / zoom));

                a[0] = cb->contact_x_max;
                a[1] = cb->contact_y_max;
                a[2] = 0.0;
                MatrixRotateHeading3(a, theta, r);
                pt[1].x = (gint16)(x + (gint)(r[0] * width / zoom));
                pt[1].y = (gint16)(y - (gint)(r[1] * height / zoom));

                a[0] = cb->contact_x_max;
                a[1] = cb->contact_y_min;
                a[2] = 0.0;
                MatrixRotateHeading3(a, theta, r);
                pt[2].x = (gint16)(x + (gint)(r[0] * width / zoom));
                pt[2].y = (gint16)(y - (gint)(r[1] * height / zoom));

                a[0] = cb->contact_x_min;
                a[1] = cb->contact_y_min;
                a[2] = 0.0;
                MatrixRotateHeading3(a, theta, r);
                pt[3].x = (gint16)(x + (gint)(r[0] * width / zoom));
                pt[3].y = (gint16)(y - (gint)(r[1] * height / zoom));

		memcpy(&pt[4], &pt[0], sizeof(GdkPoint));

	        gdk_draw_lines(pixmap, gc, pt, 5);

		break;
	    }
	}


	/* Draw icon mask. */

	/* Get icon mask by the object's type. */
	switch(o->type)
	{
          case SAR_OBJ_TYPE_GROUND:
            mask = e->view_ground_mask;
            break;
          case SAR_OBJ_TYPE_HELIPAD:
            mask = e->view_helipad_mask;
            break;
          case SAR_OBJ_TYPE_RUNWAY:
            mask = NULL;
            break;
          case SAR_OBJ_TYPE_HUMAN:
            mask = e->view_human_mask;
            break;
          case SAR_OBJ_TYPE_FIRE:
            mask = e->view_fire_mask;
            break;
          case SAR_OBJ_TYPE_SMOKE:
            mask = e->view_smoke_mask;
            break;
          default:
	    mask = e->view_object_mask;
	    break;
	}
	/* Draw runway? */
	if(o->type == SAR_OBJ_TYPE_RUNWAY)
	{
	    const sar_direction_struct *dir = &o->dir;
	    gint half_length = 15;
	    GdkGCValues gcv;

	    gdk_gc_get_values(gc, &gcv);
	    gdk_gc_set_line_attributes(
		gc,
		6,
		gcv.line_style,
		gcv.cap_style,
		gcv.join_style
	    );
	    gdk_draw_line(
		pixmap, gc,
		(gint)(x + (sin(dir->heading) * half_length)),
		(gint)(y + (cos(dir->heading) * half_length)),
                (gint)(x - (sin(dir->heading) * half_length)),
                (gint)(y - (cos(dir->heading) * half_length))
	    );
            gdk_gc_set_line_attributes(
                gc,
                gcv.line_width,
                gcv.line_style,
                gcv.cap_style,
                gcv.join_style
            );
	}
	/* Draw icon mask. */
	else if(mask != NULL)
	{
	    gdk_gc_set_clip_mask(gc, mask);
	    gdk_gc_set_clip_origin(gc, x - (15 / 2), y - (15 / 2));
	    gdk_draw_rectangle(
		pixmap, gc, TRUE,
		x - (15 / 2), y - (15 / 2),
		15, 15
	    );
	    gdk_gc_set_clip_mask(gc, NULL);
	}

	/* Draw label. */
	if(display_flags & EDITOR_VIEW_DISPLAY_OBJECT_LABELS)
	{
	    gint lbearing, rbearing, swidth, ascent, descent;
	    GdkFont *font = style->font;
	    gchar *s;

	    if(o->name != NULL)
		s = g_strdup_printf("%s(#%i)", o->name, on);
	    else
		s = g_strdup_printf("#%i", on);

            gdk_string_extents(
                font, s,
                &lbearing, &rbearing, &swidth, &ascent, &descent
            );

	    /* Draw selection background? */
	    if(state == GTK_STATE_SELECTED)
		gdk_draw_rectangle(
		    pixmap, gc, TRUE,
		    x - (swidth / 2) - 1,
		    y + (15 / 2) + 2,
		    swidth + 2,
		    ascent + descent + 2
		);

	    /* Draw text. */
            gdk_draw_string(
                pixmap, font, style->text_gc[state],
                x + lbearing - (swidth / 2),
		y + ascent + (15 / 2) + 3,
                s
            );
	    g_free(s);
	}
}

/*
 *	Redraws the view.
 */
void EditorViewDraw(editor_struct *e)
{
	gint state, width, height, half_width, half_height;
	gulong display_flags;
	GdkPixmap *pixmap;
	GdkWindow *window;
	GdkGC *gc;
	GdkGCValues gcv;
	GtkStyle *style;
        GtkWidget *w;
	GtkCList *clist;

	gfloat zoom;


        if(e == NULL)
            return;

	/* View frozen? */
	if(e->view_freeze_count > 0)
	    return;

	display_flags = e->view_display_flags;
	w = e->view_da;
	window = (w != NULL) ? w->window : NULL;
	clist = (GtkCList *)e->obj_clist;
	if((window == NULL) || (clist == NULL))
	    return;

	if(!GTK_WIDGET_MAPPED(w))
	    return;

	state = GTK_WIDGET_STATE(w);
	style = gtk_widget_get_style(w);
	pixmap = e->view_pm;
	if((pixmap == NULL) || (style == NULL))
	    return;

	gc = style->fg_gc[state];
	gdk_window_get_size(pixmap, &width, &height);
	half_width = width / 2;
	half_height = height / 2;
	if((gc == NULL) || (half_width <= 0) || (half_height <= 0))
	    return;


	/* Record foreground gc values and then modify it to suit our
	 * drawing needs.
	 */
	gdk_gc_get_values(gc, &gcv);
	gdk_gc_set_function(gc, GDK_INVERT);


	/* Get zoom and begin drawing details. */
	EditorViewGetViewZoomM(e, &zoom);
	if(zoom > 0.0f)
	{
	    gint i;
	    gfloat offset_xm, offset_ym, grid_spacing_m;
	    gint grid_spacing_xw, grid_spacing_yw;
	    GList *glist;
	    editor_obj_struct *o;

	    /* Get basic values. */
	    EditorViewGetViewPositionM(e, &offset_xm, &offset_ym);
	    EditorViewGetViewGridSpacingM(e, &grid_spacing_m);

	    /* Draw background. */
	    EditorViewDrawBackground(
		e, pixmap, display_flags,
                width, height, half_width, half_height,
                state, style,
                offset_xm, offset_ym, zoom
	    );

	    /* Draw grid. */
	    grid_spacing_xw = (gint)(grid_spacing_m * width / zoom);
	    grid_spacing_yw = (gint)(grid_spacing_m * height / zoom);
	    if((grid_spacing_xw > 32) && (grid_spacing_yw >= 32))
	    {
		gint x = ((gint)(-offset_xm * width / zoom) + half_width) %
		    grid_spacing_xw;
                gint y = ((gint)(offset_ym * height / zoom) + half_height) %
                    grid_spacing_yw;
		gint x2, y2;

		for(x2 = x - grid_spacing_xw; x2 >= 0; x2 -= grid_spacing_xw)
		    gdk_draw_line(pixmap, gc, x2, 0, x2, height);
                for(x2 = x ; x2 < width; x2 += grid_spacing_xw)
                    gdk_draw_line(pixmap, gc, x2, 0, x2, height);
                for(y2 = y - grid_spacing_yw; y2 >= 0; y2 -= grid_spacing_yw)
                    gdk_draw_line(pixmap, gc, 0, y2, width, y2);
                for(y2 = y; y2 < height; y2 += grid_spacing_yw)
                    gdk_draw_line(pixmap, gc, 0, y2, width, y2);
	    }

	    /* Draw objects, skipping selected objects. */
	    glist = clist->selection;
	    for(i = 0; i < clist->rows; i++)
	    {
		if((glist != NULL) ?
		    (g_list_find(glist, (gpointer)i) != NULL) : FALSE
		)
		    continue;

		o = EDITOR_OBJ(gtk_clist_get_row_data(clist, i));
		EditorViewDrawObject(
		    e, pixmap, display_flags,
		    width, height, half_width, half_height,
		    state, style, gc,
		    offset_xm, offset_ym, zoom,
		    i, o
		);
	    }
	    /* Draw selected objects. */
	    for(; glist != NULL; glist = glist->next)
	    {
		i = (gint)glist->data;
                o = EDITOR_OBJ(gtk_clist_get_row_data(clist, i));
                EditorViewDrawObject(
                    e, pixmap, display_flags,
                    width, height, half_width, half_height,
                    GTK_STATE_SELECTED, style, gc,
                    offset_xm, offset_ym, zoom,
                    i, o
                );
	    }
	}

	/* Draw position label? */
	if(display_flags & EDITOR_VIEW_DISPLAY_POSITION_LABEL)
	{
            gint lbearing, rbearing, swidth, ascent, descent;
	    gchar *s;
            GdkFont *font = style->font;
	    gfloat cursor_xm, cursor_ym, cursor_zm;
	    gfloat	longitude_offset_deg, latitude_offset_deg,
			planet_rm, alt = 0.0f;
	    gfloat	longitude_deg, latitude_deg;
	    GtkAdjustment *adj;
	    GtkSpinButton *spin;

	    /* Get cursor position in meters. */
	    EditorViewGetCursorM(e, &cursor_xm, &cursor_ym, &cursor_zm);

	    /* Get longitude and latitude offset in degrees. */
	    spin = (GtkSpinButton *)e->hdr_gps_longitude_spin;
	    adj = (spin != NULL) ? spin->adjustment : NULL;
	    longitude_offset_deg = (adj != NULL) ? adj->value : 0.0f;
            spin = (GtkSpinButton *)e->hdr_gps_latitude_spin;
            adj = (spin != NULL) ? spin->adjustment : NULL;
            latitude_offset_deg = (adj != NULL) ? adj->value : 0.0f;
	    /* Get planet radius in meters. */
            spin = (GtkSpinButton *)e->hdr_planetary_radius_spin;
            adj = (spin != NULL) ? spin->adjustment : NULL;
            planet_rm = MAX((adj != NULL) ? adj->value : 6371000.0f, 1.0f);

	    /* Calculate longitude and latitude of cursor position in
	     * degrees.
	     */
	    SFMMToDMS(
		cursor_xm, cursor_ym,
		planet_rm, longitude_offset_deg, latitude_offset_deg,
		&longitude_deg, &latitude_deg
	    );
	    switch(EDITOR_UNITS(e))
	    {
	      case EDITOR_UNITS_METRIC:
		alt = cursor_zm;
		break;
	      case EDITOR_UNITS_METRIC_ALT_FEET:
	      case EDITOR_UNITS_ENGLISH:
		alt = (gfloat)SFMMetersToFeet(cursor_zm);
		break;
	    }
	    s = g_strdup_printf(
		"%s %s  Alt %.2f",
		SFMLongitudeToString(longitude_deg),
		SFMLatitudeToString(latitude_deg),
		alt
	    );
            gdk_string_extents(
                font, s,
                &lbearing, &rbearing, &swidth, &ascent, &descent
            );
            gdk_draw_string(
                pixmap, font, gc,
                5 + lbearing,
                5 + ascent,
                s
            );
            g_free(s);
	}

	/* Draw selection rectangle? */
	if(e->view_rect_select)
	{
	    gint x = MIN(e->view_rect_select_x0, e->view_rect_select_x1);
            gint y = MIN(e->view_rect_select_y0, e->view_rect_select_y1);
	    gint dx = ABSOLUTE(
		e->view_rect_select_x1 - e->view_rect_select_x0
	    );
            gint dy = ABSOLUTE(
                e->view_rect_select_y1 - e->view_rect_select_y0
            );
            gdk_draw_rectangle(
                pixmap, gc, FALSE,
                x, y, dx, dy
            );
	}


	/* Restore foreground gc values. */
	gdk_gc_set_function(gc, gcv.function);


	/* Draw focus? */
	if(GTK_WIDGET_HAS_FOCUS(w) && GTK_WIDGET_SENSITIVE(w))
	    gdk_draw_rectangle(
                pixmap, style->fg_gc[state], FALSE,
                0, 0, width - 1, height - 1
            );

	/* Send pixmap to window. */
	gdk_draw_pixmap(
	    window, style->fg_gc[state], pixmap,
	    0, 0, 0, 0, width, height
	);
}


/*
 *	View cursor draw timeout callback.
 *
 *	This draws the cursor directly to the view window without the
 *	modifying of the pixmap buffer.  Uses the inverted gc (when
 *	view is in focus) to simulate blinking each time this function
 *	is called.
 */
gint EditorViewCursorDrawTOCB(gpointer data)
{
        gint state, width, height;
        GdkWindow *window;
        GdkGC *gc;
        GtkStyle *style;
        GtkWidget *w;
	editor_struct *e = EDITOR(data);
        if(e == NULL)
            return(TRUE);

        w = e->view_da;
        window = (w != NULL) ? w->window : NULL;
        if(window == NULL)
            return(TRUE);

	if(!GTK_WIDGET_MAPPED(w))
	    return(TRUE);

	state = GTK_WIDGET_STATE(w);
	style = gtk_widget_get_style(w);
        gdk_window_get_size(window, &width, &height);
        if((style == NULL) || (width <= 0) || (height <= 0))
            return(TRUE);

	gc = style->fg_gc[state];
	

	/* View not frozen? */
        if(e->view_freeze_count <= 0)
        {
	    gint x = e->view_cursor_xw;
	    gint y = e->view_cursor_yw;
	    gint r = 5;
	    GdkGCValues gcv;

	    /* If view is in focus then use inverted gc. */
	    if(GTK_WIDGET_HAS_FOCUS(w))
	    {
		gdk_gc_get_values(gc, &gcv);
		gdk_gc_set_function(gc, GDK_INVERT);
	    }

	    /* If cursor is in bounds then draw it. */
	    if(((x + r) >= 0) && ((x - r) < width) &&
               ((y + r) >= 0) && ((y - r) < height)
	    )
	    {
                gdk_draw_line(window, gc, x - r, y, x + r, y);
                gdk_draw_line(window, gc, x, y - r, x, y + r);
	    }

	    /* Restore gc function as needed. */
            if(GTK_WIDGET_HAS_FOCUS(w))
		gdk_gc_set_function(gc, gcv.function);
	}

	return(TRUE);
}


/*
 *	Processes a rectangular select.
 */
gint EditorViewProcessRectangularSelect(editor_struct *e)
{
	gint objects_selected = 0;
	gint xw0, xw1, yw0, yw1, width, height;
	gint i, xw, yw;
	gfloat offset_xm, offset_ym, zoom;
	GdkWindow *window;
	GtkWidget *w;
	GtkCList *clist;


	if(e == NULL)
	    return(objects_selected);

	w = e->view_da;
	window = (w != NULL) ? w->window : NULL;
	if(window == NULL)
	    return(objects_selected);

        if(!GTK_WIDGET_MAPPED(w))
            return(objects_selected);

	gdk_window_get_size(window, &width, &height);

	clist = (GtkCList *)e->obj_clist;
	if(clist == NULL)
	    return(objects_selected);

	/* Get rectangular select bounds in window coordinates. */
	xw0 = MIN(e->view_rect_select_x0, e->view_rect_select_x1);
        xw1 = MAX(e->view_rect_select_x0, e->view_rect_select_x1);
        yw0 = MIN(e->view_rect_select_y0, e->view_rect_select_y1);
        yw1 = MAX(e->view_rect_select_y0, e->view_rect_select_y1);

	EditorViewGetViewPositionM(e, &offset_xm, &offset_ym);
	EditorViewGetViewZoomM(e, &zoom);
	if(zoom <= 0.0f)
	    return(objects_selected);

	/* Begin selecting. */
        gtk_clist_freeze(clist);
	e->view_freeze_count++;

	/* Unselect all? */
	if(!(e->drag_key_state & GDK_CONTROL_MASK))
	    gtk_clist_unselect_all(clist);

	/* Iterate through all objects. */
	for(i = clist->rows - 1; i >= 0; i--)
	{
	    editor_obj_struct *o = EDITOR_OBJ(
		gtk_clist_get_row_data(clist, i)
	    );
	    if(o == NULL)
		continue;

	    /* Convert object position into window coordinates. */
	    EditorViewMToW(
		(gfloat)o->pos.x, (gfloat)o->pos.y,
		offset_xm, offset_ym, zoom,
		&xw, &yw,
		width, height
	    );
	    /* Check if the object's position in window coordinates
	     * falls within the selection area.
	     */
	    if((xw >= xw0) && (xw <= xw1) &&
               (yw >= yw0) && (yw <= yw1)
	    )
	    {
		gtk_clist_select_row(clist, i, 0);
		objects_selected++;
	    }
	}

        e->view_freeze_count--;
	gtk_clist_thaw(clist);

	return(objects_selected);
}
